package hexagon.world.gen;

import java.util.Random;

import hexagon.world.Block;
import hexagon.world.Chunk;
import hexagon.world.World;
import hexagon.world.block.BlockType;

public class ChunkGen {
	private NoiseGenOctaves gen;
	private World world;
	
	public ChunkGen(World world, Random random) {
		this.world = world;
		gen = new NoiseGenOctaves(random, 7);
	}
	
	public void genChunk(Chunk chunk) {
		int cx = chunk.x() * 16, cy = chunk.y() * 16, cz = chunk.z() * 16;
		for (int i = 0; i < 16; i++) {
			for (int j = 0; j < 16; j++) {
				double xx = cx + i;
				double zTemp = cz + j + xx / 2.0;
				double yy = Math.sin(Math.PI * 2 * zTemp / Chunk.ringSize) * Chunk.ringSize / 4;
				double zz = Math.cos(Math.PI * 2 * zTemp / Chunk.ringSize) * Chunk.ringSize / 4;
				double scale = 0.002;
				double amplitude = 100;
				int height = (int) (gen.genNoiseOctaves(xx * scale, yy * scale, zz * scale) * amplitude) - cy;
				if (height >= 0 && height < 16) chunk.setBlock(new Block(i, height, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, chunk));
				if (height - 1 >= 0 && height - 1 < 16) chunk.setBlock(new Block(i, height - 1, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, chunk));
				if (height - 2 >= 0 && height - 2 < 16) chunk.setBlock(new Block(i, height - 2, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, chunk));
			}
		}
	}
}
