#version 330 core
#define PI 3.14159265
#define PId3 PI / 3
#define y60 0.8660254

layout (triangles) in;
layout (triangle_strip, max_vertices = 3) out;

in VertOut {
	vec2 texCoords;
	int blockTex;
	vec3 normal;
} vertOut[];

out FragIn {
	vec2 texCoords;
	vec2 blockTex;
	vec3 normal;
} fragIn;

uniform mat4 projectionMatrix;
uniform mat4 viewMatrix;
uniform float totalSize;
uniform float radius;
uniform vec3 cam;

void main() {
	vec2 blockTex = vec2(vertOut[0].blockTex & 0xffff, (vertOut[0].blockTex >> 16) & 0xffff);
	fragIn.blockTex = blockTex;
	mat4 matrix = projectionMatrix * viewMatrix;
	for (int i = 0; i < 3; i++) {
		vec4 pos = gl_in[i].gl_Position;
		pos.z = mod(mod((pos.z + totalSize / 2), totalSize) - cam.z / y60, totalSize) - totalSize / 2;
		float mult = exp((pos.y - cam.y) / radius);
		float v = pos.z * PI / totalSize;
		float z = -sin(v);
		float y = cos(v);
		
		fragIn.normal = mat3(1, 0, 0, 0, y, -z, 0, z, y) * vertOut[i].normal;
		
		float len = sqrt(z*z+y*y);
		y /= len;
		z /= len;
		y *= radius;
		z *= radius;
		pos = vec4(pos.x - cam.x, y, z, 0) * mult;
		pos.y -= radius;
		pos.w = 1;
		gl_Position = matrix * pos; // temp
		fragIn.texCoords = vertOut[i].texCoords;
		EmitVertex();
	}
	EndPrimitive();
}