package hexagon.world.gen;

import java.util.Random;

// Improved Perlin Noise: http://mrl.nyu.edu/~perlin/noise/
public class NoiseGen {
	private int[] perm;
	
	public NoiseGen(Random random) {
		perm = new int[512];
		
		for (int i = 0; i < 256; perm[i] = i++);
		
		for (int i = 0; i < 256; i++) {
			int idx = random.nextInt(256 - i) + i;
			int temp = perm[i];
			perm[i] = perm[idx];
			perm[idx] = temp;
			perm[i + 256] = perm[i];
		}
	}
	
	private double lerp(double t, double a, double b) {
		return a + t * (b - a);
	}
	
	private double grad(int hash, double x, double y, double z) {
		int h = hash & 15;
		double u = h < 8 ? x : y,
				v = h < 4 ? y : h == 12 || h == 14 ? x : z;
		return ((h & 1) == 0 ? u : -u) + ((h & 2) == 0 ? v : -v);
	}
	
	double fade(double t) {
		return t * t * t * (t * (t * 6 - 15) + 10);
	}
	
	public double noise(double x, double y, double z) {
		int ix = (int) Math.floor(x);
		int iy = (int) Math.floor(y);
		int iz = (int) Math.floor(z);
		x -= ix;
		y -= iy;
		z -= iz;
		ix &= 255;
		iy &= 255;
		iz &= 255;
		double u = fade(x);
		double v = fade(y);
		double w = fade(z);
		
		int A = perm[ix] + iy, AA = perm[A] + iz, AB = perm[A + 1] + iz;
		int B = perm[ix + 1] + iy, BA = perm[B] + iz, BB = perm[B + 1] + iz;
		
		return	lerp(w, lerp(v, lerp(u, grad(perm[AA  ], x  , y  , z   ),
										grad(perm[BA  ], x-1, y  , z   )),
								lerp(u, grad(perm[AB  ], x  , y-1, z   ),
										grad(perm[BB  ], x-1, y-1, z   ))),
						lerp(v, lerp(u, grad(perm[AA+1], x  , y  , z-1 ),
										grad(perm[BA+1], x-1, y  , z-1 )),
								lerp(u, grad(perm[AB+1], x  , y-1, z-1 ),
										grad(perm[BB+1], x-1, y-1, z-1 ))));
	}
}
