package hexagon;

import java.util.function.Consumer;

import org.lwjgl.input.Keyboard;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;

import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.util.MathUtil;
import hexagon.util.Vec3f;
import hexagon.world.World;

public class Camera {
	
	public float fov, aspect, near, far, camX, camY;
	public Vector3f position;
	public Vec3f realPosition;
	public Vector3f rotation;
	public Vector3f blockCoords, placeInBlock;
	public Matrix4f projectionMatrix, viewMatrix;
	
	public boolean fly = true, F_down;
	public Vector3f playerVelocity;
	private World world;
	
	private Consumer<Shader> updateUniforms = s -> {
		s.setUniformMat4("viewMatrix", viewMatrix);
		s.setUniform1f("camX", camX);
		s.setUniform1f("camY", camY);
	};
	private Consumer<Shader> setProjMatrix = s -> s.setUniformMat4("projectionMatrix", projectionMatrix);
	
	public Camera(float fov, float aspect, float near, float far, World world) {
		this.fov = fov;
		this.aspect = aspect;
		this.near = (float) (near * World.xMult * 12);
		this.far = (float) (far * World.xMult * 12);
		this.world = world;
		position = new Vector3f();
		realPosition = new Vec3f();
		blockCoords = new Vector3f();
		placeInBlock = new Vector3f();
		rotation = new Vector3f();
		playerVelocity = new Vector3f();
		projectionMatrix = MathUtil.projectionMatrix(fov, aspect, this.near, this.far, null);
		viewMatrix = MathUtil.viewMatrix(this, null);
	}
	
	public void update() {
		float speed = Keyboard.isKeyDown(Keyboard.KEY_LCONTROL) ? 0.01f : Keyboard.isKeyDown(Keyboard.KEY_RCONTROL) ? 10f : 0.1f, rSpeed = 0.05f;
		float xSpeed = speed;//(float) (speed * World.radius / Chunk.totalSize);
		
		if (Keyboard.isKeyDown(Keyboard.KEY_W)) {
			playerVelocity.z += (float) (Math.cos(rotation.y) * speed);
			playerVelocity.x += (float) (Math.sin(rotation.y) * xSpeed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_S)) {
			playerVelocity.z -= (float) (Math.cos(rotation.y) * speed);
			playerVelocity.x -= (float) (Math.sin(rotation.y) * xSpeed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_D)) {
			playerVelocity.x += (float) (Math.cos(rotation.y) * xSpeed);
			playerVelocity.z -= (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_A)) {
			playerVelocity.x -= (float) (Math.cos(rotation.y) * xSpeed);
			playerVelocity.z += (float) (Math.sin(rotation.y) * speed);
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_SPACE)) {
			if (fly) position.y += speed;
			else if (playerVelocity.y == 0) playerVelocity.y += speed;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_LSHIFT)) {
			if (fly) position.y -= speed;
		}
		
		float velocityFactor = (float) Math.hypot(playerVelocity.x, playerVelocity.z) / speed;
		if (velocityFactor > 1) {
			playerVelocity.x /= velocityFactor;
			playerVelocity.z /= velocityFactor;
		}
		if (!fly) {
			playerVelocity.y -= 9.82f / 3600;
			
			if (placeInBlock.y > -0.2f && world.getBlock((int) blockCoords.x, (int) blockCoords.y - 2, (int) blockCoords.z) != null) {
				if (playerVelocity.y < 0) playerVelocity.y = 0;
			}
		}
		
		position.x += playerVelocity.x;
		if (!fly) position.y += playerVelocity.y;
		position.z += playerVelocity.z;
		
		playerVelocity.x *= 0.75f;
		playerVelocity.z *= 0.75f;
		
		if (Keyboard.isKeyDown(Keyboard.KEY_UP)) rotation.x -= rSpeed;
		if (Keyboard.isKeyDown(Keyboard.KEY_DOWN)) rotation.x += rSpeed;
		if (Keyboard.isKeyDown(Keyboard.KEY_LEFT)) rotation.y -= rSpeed;
		if (Keyboard.isKeyDown(Keyboard.KEY_RIGHT)) rotation.y += rSpeed;
		if (Keyboard.isKeyDown(Keyboard.KEY_PRIOR)) rotation.z -= rSpeed;
		if (Keyboard.isKeyDown(Keyboard.KEY_NEXT)) rotation.z += rSpeed;
		
		if (Keyboard.isKeyDown(Keyboard.KEY_F)) {
			if (!F_down) {
				F_down = true;
				fly = !fly;
			}
		} else F_down = false;
		
		camX = (float) (position.x * World.xMult);
		camY = position.y - (float) World.radius;
		realPosition.setCyl((float) (position.x * World.xMult), World.radius, position.z / HexagonModel.y60 * World.angleHalfHexagon);
		MathUtil.toBlockCoords(position, blockCoords, placeInBlock);
		MathUtil.viewMatrix(this, viewMatrix);
		Shader.forEachShader(updateUniforms);
	}
	
	public void setPosition(float x, float y, float z) {
		position.x = x;
		position.y = y;
		position.z = z;
		playerVelocity.set(0, 0, 0);
	}
	
	public void loadProjMatrToShader(Shader shader) {
		shader.enable();
		setProjMatrix.accept(shader);
	}
	
}
