package hexagon;

import java.awt.AWTException;
import java.awt.Robot;
import java.io.File;

import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.ContextAttribs;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.PixelFormat;

import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.model.Loader;
import hexagon.util.MousePicker;
import hexagon.world.Block;
import hexagon.world.Chunk;
import hexagon.world.World;
import hexagon.world.block.BlockType;

public class Main {
	
	public static final int width = 854, height = 480;
	public static MousePicker mousePicker;
	public static Main main;
	
	private World world;
	public Camera camera;
	
	private boolean F3_down = false, L_down = false, B_down = false, M_down = false;
	public static int sides = 0xff; // Sides to render, used for debugging
	
	public Main() {
		main = this;
		initGL();
		initGame();
	}
	
	private void initGame() {
		HexagonModel.init();
		Shader.init();
		
		world = new World();
		world.setSaveFolder(new File(System.getenv("appdata") + "/Hexagon/saves/TempWorld/"));
		
		camera = new Camera(70f, 16f / 9f, 0.02f, 1000f, world);
		camera.position.y = 2.7f;
		camera.loadProjMatrToShader(Shader.block);
		Shader.block.setUniform1i("blocksPerRow", 256 / 32);
		Shader.block.setUniform1f("radius", (float) World.radius);
		Shader.block.setUniform1f("totalSize", Chunk.totalSize);
		mousePicker = new MousePicker(camera);
	}
	
	private void tick() {
		System.out.println(camera.position.z * (float) Math.PI / Chunk.totalSize);
		if (Keyboard.isKeyDown(Keyboard.KEY_F3)) {
			if (!F3_down) {
				F3_down = true;
				Shader.init();
				camera.loadProjMatrToShader(Shader.block);
				Shader.block.setUniform1f("radius", (float) World.radius);
				Shader.block.setUniform1f("totalSize", Chunk.totalSize);
				Loader.theLoader.reloadTextures();
				//HexagonModel.init();
				System.out.println("Reloaded textures and shaders.");
			}
		} else F3_down = false;
		if (Keyboard.isKeyDown(Keyboard.KEY_DELETE)) {
			if (Keyboard.isKeyDown(Keyboard.KEY_P)) {
				camera.setPosition(0, 2.7f, 0);
			}
			if (Keyboard.isKeyDown(Keyboard.KEY_R)) {
				camera.rotation.x = camera.rotation.y = camera.rotation.z = 0;
			}
			if (Keyboard.isKeyDown(Keyboard.KEY_L)) {
				if (!L_down) {
					L_down = true;
					world.unloadWorld();
					world = new World();
				}
			} else L_down = false;
		}
		if (Keyboard.isKeyDown(Keyboard.KEY_B)) {
			if (!B_down) {
				B_down = true;
				int x = (int) camera.blockCoords.x, y = (int) camera.blockCoords.y, z = (int) camera.blockCoords.z;
				world.setBlock(new Block(x, y, z, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, world.getChunkForSure(x >> 4, y >> 4, z >> 4)));
			}
		} else B_down = false;
		if (Keyboard.isKeyDown(Keyboard.KEY_M)) {
			if (!M_down) {
				M_down = true;
				camera.rotation.x = (float) Math.PI / 2;
				try {
					Robot robot = new Robot();
					int x1 = Mouse.getX(), y1 = Mouse.getY();
					robot.mouseMove(Display.getX() + Mouse.getX(), Display.getY() + Display.getHeight() - Mouse.getY());
					Display.update();
					x1 = Mouse.getX() - x1;
					y1 = Mouse.getY() - y1;
					robot.mouseMove(Display.getX() + x1 + Display.getWidth() / 2, Display.getY() + y1 + Display.getHeight() / 2);
				} catch (AWTException e) {
					e.printStackTrace();
				}
			}
		} else M_down = false;
		
		if (Keyboard.isKeyDown(Keyboard.KEY_0)) sides |= 1;
		if (Keyboard.isKeyDown(Keyboard.KEY_1)) sides |= 1 << 1;
		if (Keyboard.isKeyDown(Keyboard.KEY_2)) sides |= 1 << 2;
		if (Keyboard.isKeyDown(Keyboard.KEY_3)) sides |= 1 << 3;
		if (Keyboard.isKeyDown(Keyboard.KEY_4)) sides |= 1 << 4;
		if (Keyboard.isKeyDown(Keyboard.KEY_5)) sides |= 1 << 5;
		if (Keyboard.isKeyDown(Keyboard.KEY_6)) sides |= 1 << 6;
		if (Keyboard.isKeyDown(Keyboard.KEY_7)) sides |= 1 << 7;
		if (Keyboard.isKeyDown(Keyboard.KEY_8)) sides = 0xff;
		if (Keyboard.isKeyDown(Keyboard.KEY_9)) sides = 0;
		if (Keyboard.isKeyDown(Keyboard.KEY_EQUALS)) World.viewDistance += 0.01f;
		if (Keyboard.isKeyDown(Keyboard.KEY_MINUS)) World.viewDistance -= 0.01f;
		
		camera.update();
		mousePicker.update();
		world.tick(camera);
	}
	
	private void loop() {
		long prevTime = System.nanoTime(), currentTime, realPrevTime = prevTime, realDeltaTime = 0;
		int delta, ticks = 0, frames = 0, fps = 0;
		while (!Display.isCloseRequested()) {
			delta = (int) (((currentTime = System.nanoTime()) - prevTime) * 1e-9 * 60);
			realPrevTime = currentTime;
			for (int i = 0; i < delta; i++) {
				tick();
				world.updateLoadedChunks(camera);
				ticks++;
				if (ticks % 60 == 0) {
					fps = frames;
					frames = 0;
				}
				prevTime += 1e9 / 60;
			}
			
			GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
			world.render(camera);
			
			int error = GL11.glGetError();
			if (error != GL11.GL_NO_ERROR) System.out.println("OpenGL error: " + error);
			
			String debugInfo = "    C: (" + camera.position.x + ", " + camera.position.y + ", " + camera.position.z + ")    E: (" + camera.realPosition.getEucl().x + ", " + camera.realPosition.getEucl().y + ", " + camera.realPosition.getEucl().z + ")    R: (" + (float) (camera.rotation.x * 180 / Math.PI) + ", " + (float) (camera.rotation.y * 180 / Math.PI) + ", " + (float) (camera.rotation.z * 180 / Math.PI) + ")    v=" + (int) (100 * World.viewDistance) / 100f;
			Display.update();
			realDeltaTime = System.nanoTime() - realPrevTime;
			frames++;
			Display.setTitle(fps + " fps   ms: " + ((int) (realDeltaTime * 1e-6) < 10 ? "0" : "") + (int) (realDeltaTime * 1e-6) + debugInfo);
			Display.sync(60);
		}
	}
	
	private void initGL() {
		try {
			Display.setDisplayMode(new DisplayMode(width, height));
			Display.create(new PixelFormat().withSamples(0).withDepthBits(24), new ContextAttribs(3, 3));
			Keyboard.create();
			Mouse.create();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		GL11.glClearColor(0, 0, 0, 1);
		GL11.glEnable(GL13.GL_MULTISAMPLE);
		GL11.glEnable(GL11.GL_DEPTH_TEST);
		GL11.glDepthFunc(GL11.GL_LEQUAL);
	}
	
	public static void destroy() {
		Loader.theLoader.cleanUp();
		Shader.cleanUp();
		Keyboard.destroy();
		Mouse.destroy();
		Display.destroy();
	}
	
	public static void main(String[] args) {
		File file = new File("lib/natives");
		if (!file.exists()) file = new File("lib/natives-" + (System.getProperty("os.name").toLowerCase().contains("win") ? "windows" : "linux"));
		System.setProperty("org.lwjgl.librarypath", file.getAbsolutePath());
		
		new Main().loop();
		destroy();
	}
	
}
