package hexagon.world;

import java.util.Arrays;

import hexagon.world.block.BlockType;

public class Chunk {
	/** Number of blocks in a chunk in one dimension (eg. 4 means 4x4x4 blocks).<br><b>Do not change!</b> */
	public static final int chunkSize = 16;
	/** Number of chunks in a ring */
	public static final int ringSize = 256;//256k works, 64k acceptable, 16k good, 4k excellent
	public static final int totalSize = chunkSize * ringSize;
	
	protected World world;
	protected final Block[] blocks = new Block[16 * 16 * 16];
	public final int x, y, z;
	
	public Chunk(int x, int y, int z, World world) {
		this.x = x;
		this.y = y;
		this.z = z;
		this.world = world;
		
		if (!loadFromFile()) {
			if (y == -1) {// the rest is not implemented yet!
				for (int i = 0; i < Chunk.chunkSize; i++) {
					for (int j = 0; j < Chunk.chunkSize; j++) {
						for (int k = 10; k < 14; k++) {
							setBlock(new Block(i, k, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
						}
						if (Math.random() < 0.9) setBlock(new Block(i, 14, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
						if (Math.random() < 0.8) setBlock(new Block(i, 15, j, Math.random() < 0.8 ? BlockType.grass : BlockType.sand, this));
					}
				}
			}
		}
	}
	
	public Block getBlock(short block) {
		return blocks[block];
	}
	
	public Block getBlock(int x, int y, int z) {
		return blocks[encodeCoords(x, y, z)];
	}
	
	public Block setBlock(Block block) {
		long coords = World.encodeBlockCoords(block.x(), block.y(), block.z());
		world.addBlockToUpdateList(coords, block);
		return blocks[(short) (coords & 0xfff)] = block;
	}
	
	public Block removeBlock(int x, int y, int z) {
		short coords = encodeCoords(x, y, z);
		long wCoords = World.encodeBlockCoords(this.x, this.y, this.z, coords);
		Block b = blocks[coords];
		blocks[coords] = null;
		
		if (b.sidesToRender != 0) {
			for (int i = 0; i < 8; i++) {
				if ((b.sidesToRender & 1 << i) != 0)
					world.removeFromRenderList(i, wCoords);
			}
		}
		
		b.onRemove(false);
		
		return b;
	}
	
	public void onRemove(boolean noUpdate) {
		saveChunk();
		
		long thisChunk = World.encodeCoords(x, y, z);
		Block b;
		for (short coords = 0; coords < blocks.length; coords++) {
			if ((b = blocks[coords]) == null) continue;
			
			if (b.sidesToRender != 0) {
				for (int i = 0; i < 8; i++) {
					if ((b.sidesToRender & 1 << i) != 0)
						world.removeFromRenderList(i, World.encodeBlockCoords(thisChunk, coords));
				}
			}
			
			b.onRemove(false);
		}
		Arrays.fill(blocks, null);
		world = null;
	}
	
	public static short encodeCoords(int x, int y, int z) { // bits in short: 4x 4y 4z
		return (short) ((x & 0xf) << 8 | (y & 0xf) << 4 | z & 0xf);
	}
	
	protected void saveChunk() {
/*		if (world.saveFolder != null) {
			world.bytes = 0;
			world.time = 0;
			long start = 0;
			try {
				File file = new File(world.saveFolder.getAbsolutePath() + "/chunks/");
				file.mkdirs();
				file = new File(file, x + "." + z + ".dat");
				if (file.exists()) file.delete();
				file.createNewFile();
				RandomAccessFile rFile = new RandomAccessFile(file, "rw");
				FileChannel fc = rFile.getChannel();
				
				/*for (BlockColumn col : blockCols) {
					col.saveColumn(world.saveBuffer, fc);
				}*/
/*		long t2 = 0;
		for (int j = 0; j < 1000; j++) {
			for (int i = 0; i < 256; i++) {
				world.saveBuffer.putInt(i);
			}
			t2 = System.nanoTime();
			world.saveBufferFull.accept(fc);
			start += System.nanoTime() - t2;
		}
		start -= world.time;
		
		long t1 = System.nanoTime();
		rFile.close();
		System.out.println(System.nanoTime() - t1);
	} catch (IOException e) {
		e.printStackTrace();
	}
	
	System.out.println(world.bytes + " bytes written in " + start + " ns (" + world.time + " ns). That's " + (int) (world.bytes / 8e3 / (start / 1e9) * 10) / 10f + " kB/s (" + (int) (world.bytes / 8e3 / (world.time / 1e9) * 10) / 10f + " kB/s)");
	
	}*/
	}
	
	/** Tries to load from file. @return Whether or not the file existed and the loading could be performed. */
	private boolean loadFromFile() {
/*		if (world.saveFolder != null) {
			world.bytes = 0;
			long start = System.nanoTime();
			try {
				File file = new File(world.saveFolder.getAbsolutePath() + "/chunks/" + x + "." + z + ".dat");
				if (!file.exists()) return false;
				file.createNewFile();
				RandomAccessFile rFile = new RandomAccessFile(file, "rw");
				FileChannel fc = rFile.getChannel();
				
					for (BlockColumn col : blockCols) {
						col.saveColumn(world.saveBuffer, fc);
					}
					world.saveBufferFull.accept(fc);
				
				rFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			start = System.nanoTime() - start;
			System.out.println(world.bytes + " bytes read in " + start + " ns.");
			
			return true;
		}*/
		return false;
	}
	
}
