package hexagon.util;

import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.Display;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

import hexagon.Camera;
import hexagon.Main;
import hexagon.model.HexagonModel;
import hexagon.world.Block;
import hexagon.world.World;

public class MousePicker {
	
	private Vector4f tempVec4;
	private Vector3f currentRay;
	private Block hoverBlock;
	private int hoverSide;
	
	private Matrix4f projectionMatrix, invProjMatr;
	private Matrix4f viewMatrix, invViewMatr;
	
	public MousePicker(Camera camera) {
		this.projectionMatrix = camera.projectionMatrix;
		viewMatrix = camera.viewMatrix;
		tempVec4 = new Vector4f();
		currentRay = new Vector3f();
	}
	
	public Vector3f getCurrentRay() {
		return currentRay;
	}
	
	public Block getHoverBlock() {
		return hoverBlock;
	}
	
	public int getHoverSide() {
		return hoverSide;
	}
	
	public void update() {
		invProjMatr = Matrix4f.invert(projectionMatrix, invProjMatr);
		invViewMatr = Matrix4f.invert(viewMatrix, invViewMatr);
		currentRay = calculateMouseRay();
	}
	
	private Vector3f calculateMouseRay() {
		if (Main.main.camera.moveWithMouse) tempVec4.set(0, 0, -1f, 1f);
		else tempVec4.set(Mouse.getX() * 2f / Display.getWidth() - 1, Mouse.getY() * 2f / Display.getHeight() - 1, -1f, 1f);
		return toWorldCoords(toEyeCoords(tempVec4));
	}
	
	private Vector3f toWorldCoords(Vector4f eyeCoords) {
		Matrix4f.transform(invViewMatr, eyeCoords, eyeCoords);
		currentRay.set(eyeCoords.x, eyeCoords.y, eyeCoords.z);
		currentRay.normalise();
		return currentRay;
	}
	
	private Vector4f toEyeCoords(Vector4f clipCoords) {// Reusing clipCoords a lot
		Matrix4f.transform(invProjMatr, clipCoords, clipCoords);
		clipCoords.set(clipCoords.x, clipCoords.y, -1f, 0f);
		return clipCoords;
	}
	
	public void updateBlockPointedAt(World world) {
		Vector3f camEucl = Main.main.camera.realPosition.getEucl();
		Vector3f current = new Vector3f(Main.main.camera.blockCoords);
		Block block = world.getBlock((int) current.x, (int) current.y, (int) current.z);
		if (block != null) {
			hoverBlock = block;
			hoverSide = -1;
			return;
		}
		Vector3f[] points = new Vector3f[12];
		for (int i = 0; i < 12; i++)
			points[i] = new Vector3f();
		Vector3f PA = new Vector3f(), PB = new Vector3f(), temp = new Vector3f();
		Vec3f tempVec3f = new Vec3f();
		do {
			for (int i = 0; i < 12; i++) {
				double v = i % 6 / 3d * Math.PI;
				temp.set((float) Math.cos(v), 1 - i / 6, (float) (Math.sin(v) / HexagonModel.y60));
				points[i].set(MathUtil.fromBlockCoords(current, temp, tempVec3f).getEucl());
			}
			int index = 0, side = -1;
			Vector3f.sub(points[index + 6], camEucl, PA);
			Vector3f.sub(points[index], camEucl, PB);
			boolean startRight = toTheRight(currentRay, PA, PB);
			if (startRight) {
				index = 6;
				do {
					index--;
					if (index == 0) break;
					Vector3f.sub(points[index + 6], camEucl, PA);
					Vector3f.sub(points[index], camEucl, PB);
				} while (toTheRight(currentRay, PA, PB));
			} else {
				do {
					index++;
					if (index == 6) {
						System.err.println("Index == 6. This is a bug!\n\tcurrent: " + current + "\n\tcurrentRay: " + currentRay);
						return;
					}
					Vector3f.sub(points[index + 6], camEucl, PA);
					Vector3f.sub(points[index], camEucl, PB);
				} while (!toTheRight(currentRay, PA, PB));
				index--;
			}
			Vector3f.sub(points[(index + 1) % 6], camEucl, PA);
			Vector3f.sub(points[index], camEucl, PB);
			if (toTheRight(currentRay, PA, PB)) { // under ceiling
				Vector3f.sub(points[index + 6], camEucl, PA);
				Vector3f.sub(points[(index + 1) % 6 + 6], camEucl, PB);
				if (toTheRight(currentRay, PA, PB)) side = index + 2; // over floor
				else side = 1;
			} else side = 0;
			
			if (side == 0) {
				Vector3f.sub(points[(index + 5) % 6], points[index], PA);
				Vector3f.sub(points[(index + 1) % 6], points[index], PB);
			} else if (side == 1) {
				Vector3f.sub(points[(index + 1) % 6 + 6], points[index + 6], PA);
				Vector3f.sub(points[(index + 5) % 6 + 6], points[index + 6], PB);
			} else {
				Vector3f.sub(points[index], points[index + 6], PA);
				Vector3f.sub(points[(index + 1) % 6 + 6], points[index + 6], PB);
			}
			Vector3f pointOnSide = points[side == 0 ? index : index + 6];
			Vector3f normal = Vector3f.cross(PA, PB, null);
			
			float distance = Vector3f.dot(Vector3f.sub(pointOnSide, camEucl, null), normal) / Vector3f.dot(currentRay, normal);
			if (distance <= World.blockToCoordRatio * World.reachDistance) {
				byte[] offsets = Block.neighbourOffsets[side];
				current.set(current.x + offsets[0], current.y + offsets[1], current.z + offsets[2]);
				block = world.getBlock((int) current.x, (int) current.y, (int) current.z);
				if (block != null) {
					hoverBlock = block;
					hoverSide = side < 2 ? 1 - side : (side + 1) % 6 + 2; // (side - 2 + 3) % 6 + 2
					return;
				}
				continue;
			}
			break;
		} while (true);
		hoverBlock = null;
		return;
	}
	
	private boolean toTheRight(Vector3f ray, Vector3f PA, Vector3f PB) {
		Vector3f cross = Vector3f.cross(PB, ray, null);
		float dot = Vector3f.dot(PA, cross);
		return dot > 0;
	}
	
}
