package hexagon.model;

import java.awt.image.BufferedImage;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.HashMap;

import javax.imageio.ImageIO;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.EXTTextureFilterAnisotropic;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.GLContext;

public class Loader {
	
	public static Loader theLoader = new Loader();
	
	public ArrayList<Integer> vaos = new ArrayList<Integer>();
	public ArrayList<Integer> vbos = new ArrayList<Integer>();
	
	private HashMap<String, Integer> texIDs = new HashMap<String, Integer>();
	private HashMap<String, int[]> texArrayIDs = new HashMap<String, int[]>();
	
	private Loader() {
		if (theLoader != null) theLoader = this;
	}
	
	public RawModel loadToVAO(float[] positions, float[] texCoords, int[] indices) {
		int vaoID = createVAO();
		loadIBO(indices);
		storeData(0, 3, positions);
		storeData(1, 2, texCoords);
		unbindVAO();
		return new RawModel(vaoID, indices.length);
	}
	
	public RawModel loadToVAO(float[] positions) {
		int vaoID = createVAO();
		storeData(0, 2, positions);
		unbindVAO();
		return new RawModel(vaoID, positions.length / 2);
	}
	
	private int createVAO() {
		int vaoID = GL30.glGenVertexArrays();
		vaos.add(vaoID);
		GL30.glBindVertexArray(vaoID);
		return vaoID;
	}
	
	public int loadTextureArray(String name, int texSize) {
		int[] id = texArrayIDs.get(name);
		if (id != null) return id[0];
		
		return loadTextureArrayNoLookup(name, true, texSize);
	}
	
	private int loadTextureArrayNoLookup(String name, boolean addInLists, int texSize) {
		BufferedImage image = null;
		try {
			image = ImageIO.read(new FileInputStream("res/" + name + ".png"));
		} catch (IOException e) {
			e.printStackTrace();
			return -1;
		}
		int width = image.getWidth(), height = image.getHeight();
		int[] pix = image.getRGB(0, 0, width, height, null, 0, width);
		ByteBuffer buf = BufferUtils.createByteBuffer(pix.length * 4);
		for (int y = 0, yLen = height / texSize; y < yLen; y++) {
			for (int x = 0, xLen = width / texSize; x < xLen; x++) {
				for (int j = 0; j < texSize; j++) {
					for (int i = 0; i < texSize; i++) {
						int idx = x * texSize + i + (y * texSize + j) * width;
						buf.put((byte) (pix[idx] >> 16 & 0xff));
						buf.put((byte) (pix[idx] >> 8 & 0xff));
						buf.put((byte) (pix[idx] >> 0 & 0xff));
						buf.put((byte) (pix[idx] >> 24 & 0xff));
					}
				}
			}
		}
		buf.flip();
		int texID = GL11.glGenTextures();
		GL11.glBindTexture(GL30.GL_TEXTURE_2D_ARRAY, texID);
		GL12.glTexImage3D(GL30.GL_TEXTURE_2D_ARRAY, 0, GL11.GL_RGBA, texSize, texSize, width / texSize * height / texSize, 0, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buf);
		if (GLContext.getCapabilities().GL_EXT_texture_filter_anisotropic) {
			float amt = Math.min(16f, GL11.glGetFloat(EXTTextureFilterAnisotropic.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT));
//			GL11.glTexParameterf(GL30.GL_TEXTURE_2D_ARRAY, EXTTextureFilterAnisotropic.GL_TEXTURE_MAX_ANISOTROPY_EXT, amt);
		} else {
			System.out.println("Anisotropic filtering is not supported :(");
		}
		GL11.glTexParameteri(GL30.GL_TEXTURE_2D_ARRAY, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);//_MIPMAP_LINEAR);
		GL11.glTexParameteri(GL30.GL_TEXTURE_2D_ARRAY, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
		GL11.glTexParameteri(GL30.GL_TEXTURE_2D_ARRAY, GL11.GL_TEXTURE_WRAP_S, GL12.GL_CLAMP_TO_EDGE);
		GL11.glTexParameteri(GL30.GL_TEXTURE_2D_ARRAY, GL11.GL_TEXTURE_WRAP_T, GL12.GL_CLAMP_TO_EDGE);
//		GL30.glGenerateMipmap(GL30.GL_TEXTURE_2D_ARRAY);// TODO: custom mipmap-generation by 1:2 scaling using average of the 4 pixels repeated 4-5 times
		
		if (addInLists) texArrayIDs.put(name, new int[] { texID, texSize });
		return texID;
	}
	
	private int loadTextureNoLookup(String name, boolean addInLists) {
		throw new RuntimeException("Loader::loadTextureNoLookup not implemented yet!");
		/*
		BufferedImage image = null;
		try {
			image = ImageIO.read(new FileInputStream("res/" + name + ".png"));
		} catch (IOException e) {
			e.printStackTrace();
			return -1;
		}
		int[] pix = image.getRGB(0, 0, image.getWidth(), image.getHeight(), null, 0, image.getWidth());
		ByteBuffer buf = BufferUtils.createByteBuffer(pix.length * 4);
		for (int i = 0; i < pix.length; i++) {
			buf.put((byte) (pix[i] >> 16 & 0xff));
			buf.put((byte) (pix[i] >> 8 & 0xff));
			buf.put((byte) (pix[i] >> 0 & 0xff));
			buf.put((byte) (pix[i] >> 24 & 0xff));
		}
		buf.flip();
		int texID = GL11.glGenTextures();
		GL11.glBindTexture(GL11.GL_TEXTURE_2D, texID);
		GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, GL11.GL_RGBA, image.getWidth(), image.getHeight(), 0, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buf);
		//		GL30.glGenerateMipmap(GL11.GL_TEXTURE_2D);
		GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
		GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
		if (GLContext.getCapabilities().GL_EXT_texture_filter_anisotropic) {
			float amt = Math.min(4f, GL11.glGetFloat(EXTTextureFilterAnisotropic.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT));
		//			GL11.glTexParameterf(GL11.GL_TEXTURE_2D, EXTTextureFilterAnisotropic.GL_TEXTURE_MAX_ANISOTROPY_EXT, amt);
		} else {
			System.out.println("Anisotropic filtering is not supported :(");
		}
		
		if (addInLists) texArrayIDs.put(name, texID);
		return texID;
		 */
	}
	
	private void storeData(int attrNumber, int coordinateSize, float[] data) {
		int vboID = GL15.glGenBuffers();
		vbos.add(vboID);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboID);
		FloatBuffer buffer = storeInFloatBuffer(data);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, buffer, GL15.GL_STATIC_DRAW);
		GL20.glVertexAttribPointer(attrNumber, coordinateSize, GL11.GL_FLOAT, false, 0, 0);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0); // Unnescesary
		GL20.glEnableVertexAttribArray(attrNumber);
	}
	
	private void unbindVAO() {
		GL30.glBindVertexArray(0);
	}
	
	private void loadIBO(int[] ids) {
		int vbo = GL15.glGenBuffers();
		vbos.add(vbo);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, vbo);
		GL15.glBufferData(GL15.GL_ELEMENT_ARRAY_BUFFER, storeInIntBuffer(ids), GL15.GL_STATIC_DRAW);
	}
	
	private IntBuffer storeInIntBuffer(int[] data) {
		IntBuffer buffer = BufferUtils.createIntBuffer(data.length);
		buffer.put(data);
		buffer.flip();
		return buffer;
	}
	
	private FloatBuffer storeInFloatBuffer(float[] data) {
		FloatBuffer buffer = BufferUtils.createFloatBuffer(data.length);
		buffer.put(data);
		buffer.flip();
		return buffer;
	}
	
	public void cleanUp() {
		for (int i : vaos)
			GL30.glDeleteVertexArrays(i);
		for (int i : vbos)
			GL15.glDeleteBuffers(i);
		
		vaos.clear();
		vbos.clear();
	}
	
	public void reloadTextures() {
		for (String s : texIDs.keySet()) {
			int id = texIDs.get(s), newId;
			GL11.glDeleteTextures(id);
			newId = loadTextureNoLookup(s, false);
			if (newId != id) System.out.println("ID-mismatch for texture. prev: " + id + ", new: " + newId);
		}
		
		for (String s : texArrayIDs.keySet()) {
			int[] grej = texArrayIDs.get(s);
			int id = grej[0], newId;
			GL11.glDeleteTextures(id);
			newId = loadTextureArrayNoLookup(s, false, grej[1]);
			if (newId != id) System.out.println("ID-mismatch for texture. prev: " + id + ", new: " + newId);
		}
	}
	
}
