package hexagon.world;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.util.function.Consumer;

import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.util.vector.Vector3f;

import gnu.trove.iterator.TLongIterator;
import gnu.trove.list.array.TLongArrayList;
import gnu.trove.map.hash.TLongIntHashMap;
import gnu.trove.map.hash.TLongObjectHashMap;
import hexagon.Camera;
import hexagon.Main;
import hexagon.graphics.BlockRenderer;
import hexagon.graphics.Shader;
import hexagon.model.HexagonModel;
import hexagon.util.MathUtil;

public class World { // TODO: Move stuff from Chunk and Block to World, and make BlockTypes
	
	public static float viewDistance = 5f;
	public static float reachDistance = 8f;
	public static final double angleHalfHexagon = Math.PI / Chunk.totalSize;
	public static final double radius = HexagonModel.y60 / angleHalfHexagon;
	public static final float blockToCoordRatio = HexagonModel.y60 * 2;
	
	private TLongObjectHashMap<Cluster> clusters = new TLongObjectHashMap<>();
	private TLongArrayList blocksToUpdate = new TLongArrayList();
	private TLongArrayList blocksUpdating = new TLongArrayList();
	private Block hoverBlock;
	private int rightClickCountdown = 0, leftClickCountdown = 0;
	private boolean leftDownPrev, rightDownPrev;
	
	// Rendering blocks
	private int[] blockRenderStrides = new int[8];
	private ByteBuffer[] blockRenderBuffers = new ByteBuffer[8];
	private TLongIntHashMap[] blockToIndexMaps = new TLongIntHashMap[8];
	private long[][] indexToBlockLists = new long[8][10000];
	
	// Saving to file
	protected File saveFolder;
	protected ByteBuffer saveBuffer = ByteBuffer.allocate(1024);
	protected ByteBuffer readBuffer = ByteBuffer.allocate(1024);
//	protected long bytes = 0, time = 0;
	protected Consumer<FileChannel> saveBufferFull = (fc) -> {
		try {
//			bytes += saveBuffer.position();
			saveBuffer.flip();
//			long start = System.nanoTime();
			fc.write(saveBuffer);
//			time += System.nanoTime() - start;
			saveBuffer.clear();
		} catch (IOException e) {
			e.printStackTrace();
		}
	};
	
	public World() {
		for (int i = 0; i < 8; i++)
			blockRenderStrides[i] = i < 2 ? BlockRenderer.block.instanceSize : BlockRenderer.side.instanceSize;
		for (int i = 0; i < 8; i++)
			blockRenderBuffers[i] = ByteBuffer.allocate(10000 * blockRenderStrides[i]).order(ByteOrder.nativeOrder());
		for (int i = 0; i < 8; i++)
			blockToIndexMaps[i] = new TLongIntHashMap();
	}
	
	public void tick(Camera camera) {
		updateHoverBlock();
		updateBlocks();
	}
	
	private void updateBlocks() {
		TLongArrayList tempList = blocksUpdating;
		blocksUpdating = blocksToUpdate;
		blocksToUpdate = tempList;
		
		long b;
		Block block;
		for (TLongIterator it = blocksUpdating.iterator(); it.hasNext();) {
			b = it.next();
			updateBlock(b);
		}
		
		for (TLongIterator it = blocksUpdating.iterator(); it.hasNext();) {
			b = it.next();
			block = getBlock(b);
			if (block != null) block.willUpdate = false;
		}
		
		blocksUpdating.clear();
	}
	
	// TODO: Rewrite to be smarter!
	public void updateLoadedChunks(Camera camera) {
		Vector3f cVec = new Vector3f(), rVec = new Vector3f(); // chunk, real
		float camX = rVec.x = camera.position.x / 16;
		float camY = rVec.y = camera.position.y / 16;
		float camZ = rVec.z = MathUtil.superMod(camera.position.z / 16, Chunk.ringSize * 2 * HexagonModel.y60);
		MathUtil.realCoordsToChunk(rVec, cVec);
		
		if (getChunk((int) Math.floor(cVec.x), (int) Math.floor(cVec.y) - 1, (int) Math.floor(cVec.z)) == null) {
			setNewChunk((int) Math.floor(cVec.x), (int) Math.floor(cVec.y) - 1, (int) Math.floor(cVec.z));
			System.out.println("start");
		} else {
			for (int cl = 0; cl < clusters._set.length; cl++) {
				Cluster cluster = clusters.get(clusters._set[cl]);
				if (cluster == null) continue;
				
				for (int ci = 0; ci < cluster.chunks.length; ci++) {
					Chunk c = cluster.chunks[ci];
					if (c == null) continue;
					
					int cx = c.x(), cy = c.y(), cz = c.z(), ix, iy, iz;
					cVec.x = cx + 0.5f;
					cVec.y = cy + 0.5f;
					cVec.z = cz + 0.5f;
					MathUtil.chunkCoordsToReal(cVec, rVec);
					
					float xx, yy, zz, distSq;
					xx = rVec.x - camX;
					yy = rVec.y - camY;
					zz = Math.min(zz = Math.abs(rVec.z - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
					distSq = xx * xx + yy * yy + zz * zz;
					distSq /= blockToCoordRatio * blockToCoordRatio;
					//				System.out.println(distSq + "\t" + cx + "\t" + cz + "\t" + (c >> 32));
					if (distSq > (viewDistance + 1) * (viewDistance + 1)) {
						cluster.removeChunk(cx, cy, cz);
					} else {
						for (int i = 0; i < 6; i++) {
							ix = cx + (i % 3 == 0 ? i / 3 * 2 - 1 : 0);//(2 * i + 1) % 3 - 1;
							iy = cy + ((i + 1) % 3 == 0 ? (i + 1) / 3 * 2 - 3 : 0);//(2 * i + 1) % 3 - 1;
							iz = cz + ((i + 2) % 3 == 0 ? (i + 2) / 3 * 2 - 3 : 0);//(2 * i + 1) / 3 - 1;
							iz = MathUtil.fitChunkZ(iz);
							
							long iCoords = encodeCoords(ix >> 4, iy >> 4, iz >> 4);
							Cluster theCluster = clusters.get(iCoords);
							if (theCluster == null || theCluster.getChunk(ix, iy, iz) == null) {
								cVec.x = ix + 0.5f;
								cVec.y = iy + 0.5f;
								cVec.z = iz + 0.5f;
								MathUtil.chunkCoordsToReal(cVec, rVec);
								xx = rVec.x - camX;
								yy = rVec.y - camY;
								zz = Math.min(zz = Math.abs(rVec.z - camZ), Chunk.ringSize * 2 * HexagonModel.y60 - zz); // 'zz =' is important!
								distSq = xx * xx + yy * yy + zz * zz;
								distSq /= blockToCoordRatio * blockToCoordRatio;
								if (distSq <= viewDistance * viewDistance) {
									setNewChunk(ix, iy, iz);
								}
							}
						}
					}
				}
			}
		}
	}
	
	private void updateHoverBlock() {
		Main.mousePicker.updateBlockPointedAt(this);
		hoverBlock = Main.mousePicker.getHoverBlock();
		
		if (hoverBlock != null) {
			boolean left = Mouse.isButtonDown(0), right = Mouse.isButtonDown(1);
			if (left) {
				if (!right) {
					if (leftClickCountdown == 0 || !leftDownPrev) {
						removeBlock(hoverBlock.x(), hoverBlock.y(), hoverBlock.z());
						hoverBlock = null;
						leftClickCountdown = 12;
					}
				}
			} else if (right) {
				if (rightClickCountdown == 0 || !rightDownPrev) {
					int side = Main.mousePicker.getHoverSide();
					if (side != -1) {
						int xx = hoverBlock.x() + Block.neighbourOffsets[side][0];
						int yy = hoverBlock.y() + Block.neighbourOffsets[side][1];
						int zz = hoverBlock.z() + Block.neighbourOffsets[side][2];
						zz = MathUtil.superMod(zz, Chunk.totalSize);
						Block block = getBlock(xx, yy, zz);
						if (block == null) {
							Chunk ch = getChunkForSure(xx >> 4, yy >> 4, zz >> 4);
							ch.setBlock(block = new Block(xx, yy, zz, Main.main.currentType, ch));
						}
					}
					rightClickCountdown = 12;
				}
			}
			leftDownPrev = left;
			rightDownPrev = right;
		}
		
		if (leftClickCountdown > 0) leftClickCountdown--;
		if (rightClickCountdown > 0) rightClickCountdown--;
	}
	
	public void render(Camera camera) {
		Shader.block.enable();
		
		BlockRenderer<Block> blockRenderer;
//		long t1 = 0, tmp1;
		for (int i = 0; i < 8; i++) {
			if ((Main.sides & 1 << i) == 0) continue;
			Shader.block.setUniform1i("side", i);
			blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
			blockRenderer.begin();
			blockRenderer.submitPerVertexData(i);
//			tmp1 = System.nanoTime();
			
			blockRenderer.submit(blockRenderBuffers[i]);
			
//			Block b;
//			for (Chunk c : chunks.values()) {
//				for (Long l : c.blockRendered.get(i)) {
//					b = c.blocks.get(l);
//					if ((b.sidesToRender & 1 << i) == 0) System.out.println("oops!!");
//					if (b != hoverBlock) blockRenderer.submit(b);
//					num++;
//				}
//			}
//			t1 += System.nanoTime() - tmp1;
//			blockRenderer.end();
//			blockRenderer.render();
		}
//		System.out.println(t1);
		
		if (hoverBlock != null) {
			Shader.block.setUniform1f("glow", 1.5f);
			for (int i = 0; i < 8; i++) {
				Shader.block.setUniform1i("side", i);
				if ((Main.sides & 1 << i) == 0 || (hoverBlock.sidesToRender & 1 << i) == 0) continue;
				blockRenderer = i < 2 ? BlockRenderer.block : BlockRenderer.side;
				blockRenderer.begin();
				blockRenderer.submitPerVertexData(i);
				blockRenderer.submit(hoverBlock);
				blockRenderer.end();
				blockRenderer.render();
			}
			Shader.block.setUniform1f("glow", 1f);
		}
		
		if (Main.main.camera.moveWithMouse) {
			GL20.glUseProgram(0);
			
			GL30.glBindVertexArray(Main.main.camera.crosshair.getVaoID());
			GL11.glDrawArrays(GL11.GL_LINES, 0, 4);
		}
	}
	
	public void submitBlock(Block b, ByteBuffer buffer) {
		if (b == null) {
			buffer.position(buffer.position() + 16);
			return;
		}
		buffer.putInt(b.x()).putInt(b.y()).putInt(b.z());
		buffer.putInt(b.blockType.texCoords);
	}
	
	protected void addToRenderList(int list, long block) {
		int N = blockToIndexMaps[list].size();
		blockToIndexMaps[list].put(block, N);
		if (N == indexToBlockLists[list].length) {
			long[] newArr = new long[N * 3 / 2];
			System.arraycopy(indexToBlockLists[list], 0, newArr, 0, N);
			indexToBlockLists[list] = newArr;
			
			ByteBuffer newBuffer = ByteBuffer.allocate(indexToBlockLists[list].length * blockRenderStrides[list]).order(ByteOrder.nativeOrder());
			System.arraycopy(blockRenderBuffers[list].array(), 0, newBuffer.array(), 0, blockRenderBuffers[list].capacity());
			newBuffer.position(blockRenderBuffers[list].position());
			blockRenderBuffers[list] = newBuffer;
		}
		indexToBlockLists[list][N] = block;
		
		if (indexToBlockLists[list].length * blockRenderStrides[list] != blockRenderBuffers[list].capacity()) System.out.println(indexToBlockLists[list].length * blockRenderStrides[list] + "\t" + blockRenderBuffers[list].capacity() + "\tA");
		
		submitBlock(getBlock(block), blockRenderBuffers[list]);
	}
	
	protected void removeFromRenderList(int list, long block) {
		if (indexToBlockLists[list].length * blockRenderStrides[list] != blockRenderBuffers[list].capacity()) System.out.println(indexToBlockLists[list].length * blockRenderStrides[list] + "\t" + blockRenderBuffers[list].capacity());
		int index = blockToIndexMaps[list].remove(block);
		int Nm1 = blockToIndexMaps[list].size(); // N - 1
		long last = indexToBlockLists[list][Nm1];
		int stride = blockRenderStrides[list];
		if (index != Nm1) {
			indexToBlockLists[list][index] = last;
			blockToIndexMaps[list].put(last, index);
			byte[] arr = blockRenderBuffers[list].array();
			try {
				System.arraycopy(arr, Nm1 * stride, arr, index * stride, stride);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		blockRenderBuffers[list].position(blockRenderBuffers[list].position() - stride);
	}
	
	public void updateBlock(int x, int y, int z) {
		updateBlock(World.encodeBlockCoords(x >> 8, y >> 8, z >> 8, Chunk.encodeCoords(x >> 4, y >> 4, z >> 4), Chunk.encodeCoords(x, y, z)));
	}
	
	public void updateBlock(long coords) {
		Block b = getBlock(coords);
		if (b == null) return;
		byte diff = b.sidesToRender, after;
		b.onUpdate();
		diff ^= after = b.sidesToRender;
		
		if (diff != 0) {
			for (int i = 0; i < 8; i++) {
				if ((diff & 1 << i) != 0) {
					if ((after & 1 << i) != 0) addToRenderList(i, coords);
					else removeFromRenderList(i, coords);
				}
			}
		}
	}
	
	public Block getBlock(long block) {
		Chunk c = getChunk(block >>> 12);
		return c == null ? null : c.getBlock((short) (block & 0xfff));
	}
	
	public Block getBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		Chunk chunk = getChunk(x >> 4, y >> 4, z >> 4);
		return chunk == null ? null : chunk.getBlock(x, y, z);
	}
	
	public void setBlock(Block block) {
		block.chunk.setBlock(block);
	}
	
	public void removeBlock(int x, int y, int z) {
		z = MathUtil.superMod(z, Chunk.totalSize);
		
		int cx = Math.floorDiv(x, 16), cy = Math.floorDiv(y, 16), cz = z / 16;
		
		Chunk chunk = getChunk(cx, cy, cz);
		if (chunk != null) chunk.removeBlock(x, y, z);
	}
	
	public Chunk getChunk(long chunk) {
		Cluster cluster = clusters.get(chunk >> 12);
		return cluster != null ? cluster.getChunk((int) (chunk & 0xfff)) : null;
	}
	
	public Chunk getChunk(int x, int y, int z) {
		Cluster cluster = clusters.get(encodeCoords(x >> 4, y >> 4, z >> 4));
		return cluster != null ? cluster.getChunk(Chunk.encodeCoords(x, y, z)) : null;
	}
	
	public Chunk getChunkForSure(int x, int y, int z) {
		z = MathUtil.fitChunkZ(z);
		int cx = x >> 4, cy = y >> 4, cz = z >> 4;
		long coords = encodeCoords(cx, cy, cz);
		Cluster cluster = clusters.get(coords);
		if (cluster == null) {
			cluster = new Cluster(cx, cy, cz, this);
			clusters.put(coords, cluster);
		}
		Chunk chunk = cluster.getChunk(Chunk.encodeCoords(x, y, z));
		if (chunk == null) {
			chunk = new Chunk(x, y, z, cluster);
			cluster.setChunk(chunk);
		}
		
		return chunk;
	}
	
	public Chunk setNewChunk(int x, int y, int z) {
		return setNewChunk(x, y, z, false);
	}
	
	public Chunk setNewChunk(int x, int y, int z, boolean genNew) {
		z = MathUtil.fitChunkZ(z);
		int cx = x >> 4, cy = y >> 4, cz = z >> 4;
		long coords = encodeCoords(cx, cy, cz);
		Cluster cluster = clusters.get(coords);
		if (cluster == null) {
			cluster = new Cluster(cx, cy, cz, this);
			clusters.put(coords, cluster);
		}
		Chunk chunk = new Chunk(x, y, z, cluster, genNew);
		cluster.setChunk(chunk);
		return chunk;
	}
	
	public static long encodeCoords(int x, int y, int z) { // bytes in long: xxxxyyzzzz
		return (x & 0xffffL) << 24 | (y & 0xffL) << 16 | z & 0xffffL;
	}
	
	public static long encodeBlockCoords(int x, int y, int z, short chunk, short block) { // bytes in long: xxxxyyzzzzcccbbb where b is blockcoords xyz and c is chunk
		return encodeCoords(x, y, z) << 24 | chunk << 12 | block;
	}
	
	public static long encodeBlockCoords(long cluster, short chunk, short block) { // bytes in long: xxxxxyyy zzzzzbbb where b is blockcoords xyz
		return cluster << 24 | chunk << 12 | block;
	}
	
	public static long encodeBlockCoords(int x, int y, int z) { // bytes in long: xxxxxyyy zzzzzbbb where b is blockcoords xyz
		return encodeBlockCoords(x >> 8, y >> 8, MathUtil.superMod(z >> 4, Chunk.totalSize) >> 4, Cluster.encodeCoords(x >> 4, y >> 4, z >> 4), Chunk.encodeCoords(x, y, z));
	}
	
	public void addBlockToUpdateList(long coords, Block b) {
		if (b.willUpdate) return;
		blocksToUpdate.add(coords);
		b.willUpdate = true;
	}
	
	public void setSaveFolder(File folder) {
		if (!folder.exists()) {
			folder.mkdir();
		}
		saveFolder = folder;
	}
	
	public void saveWorld() {
		if (saveFolder != null) {
			clusters.forEachValue(c -> {
				c.saveChunks();
				return true;
			});
		}
	}
	
	public void unloadWorld() {
		updateBlocks();
		
		clusters.forEachValue(c -> {
			c.onRemove(true);
			return true;
		});
		
		blocksToUpdate.clear();
		
		blocksToUpdate = null;
		hoverBlock = null;
		clusters = null;
	}
	
}
